/*
 * Copyright (c) 2006 Greg Rodgers All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   
 * - Redistribution of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *    
 * - Redistribution in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *   
 * The names of Greg Rodgers, Sun Microsystems, Inc. or the names of
 * contributors may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *    
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES,
 * INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. GREG RODGERS,
 * SUN MICROSYSTEMS, INC. ("SUN"), AND SUN'S LICENSORS SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL GREG 
 * RODGERS, SUN, OR SUN'S LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT 
 * OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR
 * PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY,
 * ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF GREG
 * RODGERS OR SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *   
 * You acknowledge that this software is not designed or intended for use
 * in the design, construction, operation or maintenance of any nuclear
 * facility.
 */

package net.java.joglutils.test3ds;

import net.java.joglutils.ThreeDS.*;
import com.sun.opengl.util.texture.Texture;
import com.sun.opengl.util.texture.TextureCoords;
import com.sun.opengl.util.texture.TextureIO;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import javax.imageio.ImageIO;
import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;

public class MyModel extends Model3DS
{
    private Texture[] texture;
    private TextureCoords[] textureCoords;
    
    private int compiledList = -1;
    private boolean renderModel = true;
    private boolean centerModel = false;
    
    private int modelBoundsList = -1;
    private boolean renderModelBounds = false;
    
    private int objectBoundsList = 1;
    private boolean renderObjectBounds = false;
    
    private boolean loaded = false;
    
    // Constructor
    public MyModel()
    {
    }
    
    public boolean isLoaded()
    {
        return loaded;
    }
    
    /**
     * Load the model and associated materials, etc
     * 
     * @param glDrawable
     * @param file
     * @return
     */
    public boolean load(GLAutoDrawable glDrawable, String file) 
    {
        GL gl = glDrawable.getGL();
        return load(gl, file);
    }
    
    /**
     * Load the model and associated materials, etc
     * 
     * @param gl
     * @param file
     * @return
     */
    public boolean load(GL gl, String file) 
    {    
        System.out.println("Load Model: " + file);
        
        if (!super.load(file)) {
            System.out.println("Load Model: FAILURE");
            return false;
        } 
        
        int numMaterials = materials.size();
        
        if (numMaterials > 0) {
            System.out.println("\n    Load Materials:");            
        }
        
        texture = new Texture[numMaterials];
        for (int i=0; i<numMaterials; i++) {
            // TODO:DELETE THIS OLD LINE loadTexture(materials.get(i).strFile, i);
            // TODO:DELETE THIS OLD LINE materials.get(i).texureId = i;
            
            String subFileName = "";
            
            // If this is read from a jar file, then try to find the path relative to the model
            int index = file.lastIndexOf('/');
            if (index != -1) {
                subFileName = file.substring(0,index+1);
            } else {
                // Else, the file path of the model was not from a jar file, so check maybe it
                // was from a local file and get that path.
                index = file.lastIndexOf('\\');
                
                if (index != -1) {
                     subFileName = file.substring(0,index+1);
                }
            }
             
            if (materials.get(i).strFile.length() > 0) {
                System.out.print("        Material:  " + subFileName + materials.get(i).strFile);                
                
                URL result;
                try {
                    result = ResourceRetriever.getResourceAsUrl(subFileName + materials.get(i).strFile);
                } catch(IOException e) {
                    System.out.println(" ... failed");
                    continue;
                }

                if (result != null && !result.getPath().endsWith("/") && !result.getPath().endsWith("\\")) {
                    loadTexture(result, i);
                    materials.get(i).textureId = i;
                    System.out.println(" ... done. Texture ID: " + i);
                } else {
                    System.out.println(" ... failed (no result for material)");
                }   
            }            
        }
        
        if (numMaterials > 0) {
            System.out.println("    Load Materials: Done");            
        }
        
        System.out.println("\n    Generate Lists:");  
        compiledList = gl.glGenLists(3);
        
        System.out.println("        Model List");         
        gl.glNewList(compiledList, GL.GL_COMPILE);
            genList(gl);
        gl.glEndList();
        
        modelBoundsList = compiledList + 1;

        System.out.println("        Boundary List");                 
        gl.glNewList(modelBoundsList, GL.GL_COMPILE);
            genModelBoundsList(gl);
        gl.glEndList();
        
        objectBoundsList = compiledList + 2;
        
        System.out.println("        Object Boundary List");                         
        gl.glNewList(objectBoundsList, GL.GL_COMPILE);
            genObjectBoundsList(gl);
        gl.glEndList();
        
        System.out.println("    Generate Lists: Done");  
        
        loaded = true;

        System.out.println("Load Model: Done");

        return loaded;
    }
    
// TODO: DELETE THIS OLD ONE    
//    public void render(GLAutoDrawable gLDrawable)
//    {
//        GL gl = gLDrawable.getGL();
//        gl.glCallList(compiledList);
//    }
    
    /**
     * Render the model and/or the boundary boxes
     * 
     * @param gLDrawable
     */
    public void render(GLAutoDrawable gLDrawable) {
        GL gl = gLDrawable.getGL();
        this.render(gl);
    }
    
    /**
     * Render the model and/or the boundary boxes
     * 
     * @param gl
     */
    public void render(GL gl) {
        if (centerModel) {
            Vec3 center = loader.getCenter();
            gl.glTranslatef(-center.x, -center.y, -center.z); 
        }

        boolean isTextureEnabled = gl.glIsEnabled(GL.GL_TEXTURE_2D);
        boolean isLightingEnabled = gl.glIsEnabled(GL.GL_LIGHTING);        
        boolean isMaterialEnabled = gl.glIsEnabled(GL.GL_COLOR_MATERIAL);
        gl.glEnable(GL.GL_TEXTURE_2D);
        gl.glEnable(GL.GL_LIGHTING);
        gl.glEnable(GL.GL_COLOR_MATERIAL);  
                
        if (renderModel  &&  compiledList >= 0) {
            gl.glCallList(compiledList);
        }

        // Disabled lighting for drawing the boundary lines so they are all white (or whatever I chose)
        gl.glDisable(GL.GL_LIGHTING);
        if (renderModelBounds  &&  modelBoundsList >= 0) {
            gl.glCallList(modelBoundsList);
        }
        
        if (renderObjectBounds  &&  objectBoundsList >= 0) {
            gl.glCallList(objectBoundsList);
        }
        
        // Reset the flags back for lighting and texture
        if (isTextureEnabled) {
            gl.glEnable(GL.GL_TEXTURE_2D);
        } else {
            gl.glDisable(GL.GL_TEXTURE_2D);
        }

        if (isLightingEnabled) {
            gl.glEnable(GL.GL_LIGHTING);
        } else {
            gl.glDisable(GL.GL_LIGHTING);
        }

        if (isMaterialEnabled) {
            gl.glEnable(GL.GL_COLOR_MATERIAL);
        } else {
            gl.glDisable(GL.GL_COLOR_MATERIAL);
        }
    }    

    // TODO:DELETE - OLD VERSION DELETE...Does not take into account the need to
    // rotate (affine transform) the loaded texture, does not assume the texture
    // file is in the same path relative to the model that was loaded
//    private void loadTexture(String strFile, int id)
//    {
//        File file = new File(strFile);
//        try {
//            texture[id] = TextureIO.newTexture(file, true);
//        }
//        catch (IOException e) {
//            System.err.println(e.getMessage());
//            return;
//        }
//    }
    
    /**
     * Load a texture given by the specified URL and assign it to the texture
     * id that is passed in.
     * 
     * @param url
     * @param id
     */
    private void loadTexture(URL url, int id) {
        if ( url != null ) {
            BufferedImage bufferedImage = null;

            try {
                bufferedImage = ImageIO.read(url);
            } catch (Exception e) {
                System.out.println(" ... FAILED loading texture with exception: "+e.getMessage());
                return;
            }                

            if (bufferedImage != null) {
                // This Affine Transform is required because the image is loaded upside down otherwise
                AffineTransform tx = AffineTransform.getScaleInstance(1, -1);
                tx.translate(0, -bufferedImage.getHeight(null));
                AffineTransformOp op = new AffineTransformOp(tx, AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
                if (op==null) {
                    //System.out.println(" op null");
                }
                
                if (op!=null) {
                    bufferedImage = op.filter(bufferedImage, null);
                    texture[id] = TextureIO.newTexture(bufferedImage, true);
                }
            }

            if (texture[id] == null) {
                System.out.println(" Loading Texture FAILED");
            }
        }
    }      

    /**
     * Convert an Unsigned byte to integer
     * 
     * @param b
     * @return
     */
    public int unsignedByteToInt(byte b) {
      return (int) b & 0xFF;
    }
  
    
    /**
     * Convert integer to float
     * 
     * @param i
     * @return
     */
    public float intToFloat(int i) {
        return (float) i / 255.0f;
    }    
    
    /**
     * Generate the model display list
     * 
     * @param gLDrawable
     */
    private void genList(GLAutoDrawable gLDrawable) 
    {
        GL gl = gLDrawable.getGL();
        genList(gl);
    }
    
    
    /**
     * Generate the model display list
     * 
     * @param gl
     */
    private void genList(GL gl) {
        TextureCoords coords;
        
        // What the does this code even do?  Nothing I think.
//        for (int i=0; i<materials.size(); i++) {
//            byte pColor[] = materials.get(i).color;                         
//        }        
                       
        for (int i=0; i<objects.size(); i++) {
            Obj tempObj = objects.get(i);            
     
            if(tempObj.hasTexture) {
                // This is required to repeat textures...because some are not and so only
                // part of the model gets filled in....Might be a way to check if this is
                // required per object but I'm not sure...would need to research this.
                gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_REPEAT);
                gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_REPEAT);
                                         
                texture[tempObj.materialID].enable();

                texture[tempObj.materialID].bind();
              
                coords = texture[tempObj.materialID].getImageTexCoords();   
            }

            for (int j=0; j<tempObj.numOfFaces; j++) {

                // If the object has a texture, then do nothing till later...else
                // apply the material property to it.
                if(tempObj.hasTexture) { 
                    // nothing

                // Has no texture but has a material instead and this material is
                // the FACES material, and not the OBJECTS material ID as being used
                // incorrectly below...by specification, the materialID is associated
                // with a FACE and not an OBJECT
                } else { 
                    if (tempObj.faces[j].materialID < materials.size()) {

                        byte aColor[] = materials.get(tempObj.faces[j].materialID).color;
                        float mat[] = { intToFloat(unsignedByteToInt(aColor[0])), 
                                        intToFloat(unsignedByteToInt(aColor[1])), 
                                        intToFloat(unsignedByteToInt(aColor[2])), 1.0f };
                        gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE, mat, 1);
                        gl.glColor3ub(aColor[0], aColor[1], aColor[2]);
                    }
                }

                gl.glBegin(GL.GL_TRIANGLES);
                for (int whichVertex=0; whichVertex<3; whichVertex++) {
                    int index = tempObj.faces[j].vertIndex[whichVertex];

                    gl.glNormal3f(tempObj.normals[index].x, tempObj.normals[index].y, tempObj.normals[index].z);

                    if (tempObj.hasTexture) {
                        if (tempObj.texVerts != null) {  
                            gl.glTexCoord2f(tempObj.texVerts[index].x, tempObj.texVerts[index].y);
                        }
                    } else {     
// This is not valid...the material ID does not correspond to an OBJECT but a FACE...and is
// take care of in the section a few lines above (per 3DS specification)                        
//                            //if (materials.size() < tempObj.materialID) {
//                            if (tempObj.materialID < materials.size()) {                                
////                                byte pColor[] = materials.get(tempObj.materialID).color;
//                                // Do something with the color
////                                gl.glColor3ub(pColor[0], pColor[1], pColor[2]);
//                            }
                    }

                    gl.glVertex3f(tempObj.verts[index].x, tempObj.verts[index].y, tempObj.verts[index].z);
                }
                gl.glEnd();  
            }

            if (tempObj.hasTexture) {
                texture[tempObj.materialID].disable();
            }
        }
        
        // Try this clearing of color so it won't use the previous color
        gl.glColor3f(1.0f, 1.0f, 1.0f);        
    }     
    
    /**
     * Set the flag to draw the model
     * 
     * @param state
     */
    public void setRenderModel( boolean state ) {
        renderModel = state;
    }
    
    /**
     * Set the flag to render the boundary of the model itself
     * 
     * @param state
     */
    public void setRenderModelBounds( boolean state ) {
        renderModelBounds = state;
    }
    
    /**
     *  Set the flag to render the boundaries of the objects of the model
     * 
     * @param state
     */
    public void setRenderObjectBounds( boolean state ) {
        renderObjectBounds = state;
    }
    
    /**
     * Render the desired object of the model (specified by an id number of 
     * the object)
     * 
     * @param gl
     * @param id
     */
    public void renderBoundsOfObject(GL gl, int id) {
        if (id >=0 && id <= objects.size()) {
            if (objects.get(id).bounds != null) {
                drawBounds(gl, objects.get(id).bounds);
            }
        }
    }
    
    /**
     * Set the flag to center the model at the center of the boundary
     * 
     * @param state
     */
    public void setCenterModel( boolean state ) {
        centerModel = state;
    }
    
    /**
     * Draw the boundary of the model (the large box representing the entire 
     * model and not the object in it)
     * 
     * @param gLDrawable
     */
    private void genModelBoundsList(GLAutoDrawable gLDrawable) {
        GL gl = gLDrawable.getGL();
        drawBounds(gl, loader.getBounds());
    }
    
    /**
     * Draw the boundary of the model (the large box representing the entire 
     * model and not the object in it)
     * 
     * @param gl
     */
    private void genModelBoundsList(GL gl) {
        drawBounds(gl, loader.getBounds());
    }
    
    /**
     * Draw the boundaries over all of the objects of the model
     *
     * @param gLDrawable
     */
    private void genObjectBoundsList(GLAutoDrawable gLDrawable) {
        GL gl = gLDrawable.getGL();
        genObjectBoundsList(gl);
    }  
    
    /**
     * Draw the boundaries over all of the objects of the model
     * 
     * @param gl
     */
    private void genObjectBoundsList(GL gl) {
        for (int i=0; i<objects.size(); i++) {
            if (objects.get(i).bounds != null) {
                drawBounds(gl, objects.get(i).bounds);
            }
        }               
    }
    
    /** 
     * Returns the bounds of the object.
     * 
     * @return The Bounds object representing the objects boundary
     */
    public Bounds getBounds() {
        return loader.getBounds();
    }
    
    /** 
     * Draws the bounding box of the object using the max and min extrema 
     * points.
     *
     * @param gl
     * @param bounds
     */
    private void drawBounds(GL gl, Bounds bounds) {      
        // Front Face
        gl.glBegin(GL.GL_LINE_LOOP);
            gl.glVertex3f(bounds.min.x, bounds.min.y, bounds.min.z);
            gl.glVertex3f(bounds.max.x, bounds.min.y, bounds.min.z);
            gl.glVertex3f(bounds.max.x, bounds.max.y, bounds.min.z);
            gl.glVertex3f(bounds.min.x, bounds.max.y, bounds.min.z);
        gl.glEnd();

        // Back Face
        gl.glBegin(GL.GL_LINE_LOOP);
            gl.glVertex3f(bounds.min.x, bounds.min.y, bounds.max.z);
            gl.glVertex3f(bounds.max.x, bounds.min.y, bounds.max.z);
            gl.glVertex3f(bounds.max.x, bounds.max.y, bounds.max.z);
            gl.glVertex3f(bounds.min.x, bounds.max.y, bounds.max.z);
        gl.glEnd();            

        // Connect the corners between the front and back face.
        gl.glBegin(GL.GL_LINES);
            gl.glVertex3f(bounds.min.x, bounds.min.y, bounds.min.z);
            gl.glVertex3f(bounds.min.x, bounds.min.y, bounds.max.z);

            gl.glVertex3f(bounds.max.x, bounds.min.y, bounds.min.z);
            gl.glVertex3f(bounds.max.x, bounds.min.y, bounds.max.z);

            gl.glVertex3f(bounds.max.x, bounds.max.y, bounds.min.z);
            gl.glVertex3f(bounds.max.x, bounds.max.y, bounds.max.z);

            gl.glVertex3f(bounds.min.x, bounds.max.y, bounds.min.z);
            gl.glVertex3f(bounds.min.x, bounds.max.y, bounds.max.z);
        gl.glEnd();
    }    
    
}